import ipaddress
import json
import http.client
import logging
import socket

import findssh
from findssh.coro import get_hosts as coro_get_hosts
from findssh.threadpool import get_hosts as threadpool_get_hosts
import log


class HTTPRRFWDetector:

    PORT = 80
    TIMEOUT = 1
    FINDSSH_SERVICE = 'http'
    GCODE_REQ_TEMPLATE = "/rr_gcode?gcode=%s" #resp {"buff":250}
    GET_REPLY_PATH = "/rr_reply"
    STATUS_REQ_PATH = "/rr_status"
    #{"status":"O","heaters":[2000.0,2000.0,2000.0],"active":[0.0,0.0,0.0],"standby":[0.0,0.0,0.0],"hstat":[0,0,0],"pos":[0.000,0.000,0.000],"machine":[0.000,0.000,0.000],"sfactor":100.00,"efactor":[100.00,100.00],"babystep":0.000,"tool":-1,"probe":"1000","fanPercent":[0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0],"fanRPM":0,"homed":[0,0,0],"msgBox.mode":-1,"seq":0,"resp":""}

    def __init__(self, parent=None, profile=None):
        self.logger = logging.getLogger('app.' + __name__)
        self.ip = findssh.getLANip()
        self.parent = parent
        self.profile = profile
        self.discovered_printers = []
    
    def filter_non_rrfw_hosts(self, hosts):
        filtered = []
        for host in hosts:
            host_ip = str(host[0])
            if self.parent_stop():
                break
            if self.check_ip(host_ip):
                filtered.append(host_ip)
        return filtered 

    def check_ip(self, host): 
        self.logger.debug("Detecting host:" + host)
        connection = http.client.HTTPConnection(host, self.PORT, timeout=self.TIMEOUT)
        self.logger.debug("Connecting")
        try:
            connection.connect()
            connection.request('GET', self.STATUS_REQ_PATH)
            resp = connection.getresponse()
        except (socket.error, http.client.error) as e:
            self.logger.warning("Error on http detection request: " + str(e))
        else:
            self.logger.debug("Checking status")
            if resp.status == 200:
                self.logger.debug("Reading body")
                try:
                    answer_body = resp.read()
                except (socket.error, http.client.error) as e:
                    self.logger.warning("Error on http detection body read: " + str(e))
                else:
                    self.logger.debug("Decoding json")
                    try:
                        decoded = json.loads(answer_body)
                    except (json.JSONDecodeError, AttributeError) as e:
                        self.logger.info("RRFW detector got non json response: " + str(e))
                    else:
                        self.logger.debug("Checking status dictionary")
                        status = decoded.get("status")
                        self.logger.debug('Got status:' + str(status))
                        return status

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    #@log.log_exception
    def detect(self):
        network = findssh.netfromaddress(self.ip)
        if isinstance(network, ipaddress.IPv6Network):
            hosts = threadpool_get_hosts(network, self.PORT, self.FINDSSH_SERVICE, self.TIMEOUT)
        else:
            hosts = findssh.runner(coro_get_hosts, network, self.PORT, self.FINDSSH_SERVICE, self.TIMEOUT)
        self.logger.debug("All responded hosts:%s" % hosts)
        if self.profile:
            VID, PID = self.profile['vids_pids'][0]
        else:
            VID, PID = "XXXX", "XXXX" #dummy values only for debug mode
        discovered_printers = []
        for host in self.filter_non_rrfw_hosts(hosts):
            self.logger.debug("HOST: %s" % host)
            printer = {}
            printer['SNR'] = host
            printer['IP'] = host
            printer['VID'] = VID
            printer['PID'] = PID
            if printer not in discovered_printers:
                discovered_printers.append(printer)
        self.discovered_printers = discovered_printers


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    from pprint import pprint
    d = HTTPRRFWDetector()
    d.detect()
    pprint(d.discovered_printers)
